﻿
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using UnityEngine.Events;
using System;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using SimpleJSON;
using System.Linq;
using OldMoatGames;


/* /////////////////////////////////////////////////////////////////////////////////////////////////
based on MacGruber's utils : https://hub.virtamate.com/resources/macgruber-utils.40744/

Licensed under CC BY after EarlyAccess ended. (see https://creativecommons.org/licenses/by-sa/4.0/)

///////////////////////////////////////////////////////////////////////////////////////////////// */

namespace PluginIdea.VolumeLight
{
    public static class Utils
    {
        public const string version = "124";

        public static IEnumerator WaitTimeCall(float time, Action callback)
        {
            yield return new WaitForSeconds(time);
            if (callback != null) callback();
        }

        public static Camera CenterCamera
        {
            get
            {
                if (Utils.IsInVR())
                {
                    if (SuperController.singleton.isOpenVR)
                    {
                        return SuperController.singleton.ViveCenterCamera;
                    }
                    return SuperController.singleton.OVRCenterCamera;
                }
                else
                {
                    return SuperController.singleton.MonitorCenterCamera;
                }
            }
        }

        public static string GetGameRootDir()
        {
            string dataPath = Application.dataPath;
            return dataPath.Substring(0, dataPath.LastIndexOf('/'));
        }

        public static string GetFolderFromUrl(string url)
        {
            if (url.Contains("/"))
            {
                return url.Substring(0, url.LastIndexOf("/"));
            }

            return url;
        }

        public static string GetPath(Transform current)
        {
            if (current.parent == null)
                return "/" + current.name;
            return GetPath(current.parent) + "/" + current.name;
        }

        public static Vector2 xy(this Vector3 aVector) { return new Vector2(aVector.x, aVector.y); }
        public static Vector2 xz(this Vector3 aVector) { return new Vector2(aVector.x, aVector.z); }
        public static Vector2 yz(this Vector3 aVector) { return new Vector2(aVector.y, aVector.z); }
        public static Vector2 yx(this Vector3 aVector) { return new Vector2(aVector.y, aVector.x); }
        public static Vector2 zx(this Vector3 aVector) { return new Vector2(aVector.z, aVector.x); }
        public static Vector2 zy(this Vector3 aVector) { return new Vector2(aVector.z, aVector.y); }

        public static bool IsValid(this Plane plane)
        {
            return plane.normal.sqrMagnitude > 0.5f;
        }

        public static void SetKeywordEnabled(this Material mat, string name, bool enabled)
        {
            if (enabled) mat.EnableKeyword(name);
            else mat.DisableKeyword(name);
        }

        public struct FloatPackingPrecision { public const int High = 64; public const int Low = 8; public const int Undef = 0; }
        static int ms_FloatPackingPrecision = FloatPackingPrecision.Undef;

        // OpenGL ES 2.0 GPU (graphicsShaderLevel = 30) usually have low float precision (16 bits on fragments)
        // So we lower the float packing precision on them (8 seems fine on Adreno (TM) 220, NVIDIA Tegra 3 and on Mali-450 MP)
        // https://docs.unity3d.com/Manual/SL-DataTypesAndPrecision.html
        const int kFloatPackingHighMinShaderLevel = 35;

        public static int GetFloatPackingPrecision()
        {
            if (ms_FloatPackingPrecision == FloatPackingPrecision.Undef)
            {
                ms_FloatPackingPrecision = SystemInfo.graphicsShaderLevel >= kFloatPackingHighMinShaderLevel ? FloatPackingPrecision.High : FloatPackingPrecision.Low;
            }
            return ms_FloatPackingPrecision;
        }

        public static void LoadTexture(string url, TextureSettings settings, TextureSetCallback callback)
        {
            QueueFastLoadTexture(url, settings, callback);
        }

        private static void QueueFastLoadTexture(string url, TextureSettings settings, TextureSetCallback callback)
        {
            if (ImageLoaderThreaded.singleton == null)
                return;
            if (string.IsNullOrEmpty(url))
                return;

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = false;
            queuedImage.skipCache = false;
            queuedImage.compress = settings.compress;
            queuedImage.createMipMaps = settings.createMipMaps;
            queuedImage.isNormalMap = settings.isNormalMap;
            queuedImage.linear = settings.linearColor;
            queuedImage.createAlphaFromGrayscale = settings.createAlphaFromGrayscale;
            queuedImage.createNormalFromBump = settings.createNormalFromBump;
            queuedImage.bumpStrength = settings.bumpStrength;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
            {
                Texture2D tex = qi.tex;
                if (tex != null)
                {
                    tex.wrapMode = settings.wrapMode;
                    tex.filterMode = settings.filterMode;
                    tex.anisoLevel = settings.anisoLevel;
                }
                callback(tex);
            };
            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }

        public static void SetLayer(Transform target, int layer)
        {
            target.gameObject.layer = layer;

            if (target.childCount > 0)
            {
                for (int i = 0; i < target.childCount; i++)
                {
                    Transform child = target.GetChild(i);
                    SetLayer(child, layer);
                }
            }
        }

        public static Transform CenterCameraTrans
        {
            get
            {
                if (Utils.IsInVR())
                {
                    if (SuperController.singleton.isOpenVR)
                    {
                        return SuperController.singleton.ViveCenterCamera.transform;
                    }
                    return SuperController.singleton.OVRCenterCamera.transform;
                }
                else
                {
                    return SuperController.singleton.MonitorCenterCamera.transform;
                }
            }
        }

        public static bool IsInVR()
        {
            return SuperController.singleton.isOpenVR || SuperController.singleton.isOVR;
        }

        public static float GetMin(List<float> sourceList)
        {
            float min = sourceList[0];
            for (int x = 1; x < sourceList.Count; x++)
            {
                if (min > sourceList[x])
                {
                    min = sourceList[x];
                }
            }

            return min;
        }

        public static void GetAllEmptyLightChildsFromAtom(Atom target, List<Atom> childAtoms)
        {
            GetChildsFromAtom(target, childAtoms);
        }

        private static void GetChildsFromAtom(Atom target, List<Atom> childAtoms)
        {
            var tmpAtoms = target.GetChildren().ToList();
            if (tmpAtoms.Count > 0)
            {
                tmpAtoms.ForEach(at => {
                    bool isLight = at.reParentObject.GetComponentInChildren<Light>(true) != null;
                    if (isLight || at.type == "Empty")
                    {
                        childAtoms.Add(at);
                    }

                    if (at.GetChildren().Count > 0)
                    {
                        GetChildsFromAtom(at, childAtoms);
                    }
                });
            }
        }

        public static Transform GetScaleTransform(FreeControllerV3 controllerV3)
        {
            // scale target
            Transform tmpTrans;
            SetTransformScale setTransformScale = controllerV3.GetComponentInParent<SetTransformScale>();
            if (setTransformScale == null)
            {
                tmpTrans = controllerV3.transform.parent.Find("object/rescaleObject").transform;
            }
            else
            {
                tmpTrans = setTransformScale.transform;
            }

            return tmpTrans;
        }

        public static Atom GetAtomFromController(FreeControllerV3 controllerV3)
        {
            return controllerV3.transform.GetComponentInParent<Atom>();
        }

        public static Atom GetAtomFromInstanceID(int id)
        {
            return SuperController.singleton.GetAtoms().FirstOrDefault(x => x.GetInstanceID() == id);
        }

        public static FreeControllerV3 GetFreeControllerV3ByInstanceID(int id)
        {
            return SuperController.singleton.GetAllFreeControllers().FirstOrDefault(x => x.transform.GetInstanceID() == id);
        }

        public static FreeControllerV3 GetControllerFromAtom(Atom atom)
        {
            if (atom.mainController != null)
            {
                return atom.mainController;
            }
            else if (atom.containingAtom != null)
            {
                if (atom.containingAtom.mainController != null)
                {
                    return atom.containingAtom.mainController;
                }
            }

            return null;
        }

        public static Canvas AppendCanvasObjectUnderHeadCenter(string canvasName, bool isUICanInteractive)
        {
            Transform headCenter = SuperController.singleton.centerCameraTarget.transform;

            GameObject canvasObj = new GameObject();
            canvasObj.name = canvasName;
            var popupCanvas = canvasObj.AddComponent<Canvas>();
            popupCanvas.renderMode = RenderMode.WorldSpace;
            CanvasScaler cs = canvasObj.AddComponent<CanvasScaler>();
            cs.scaleFactor = 100.0f;
            cs.dynamicPixelsPerUnit = 4f;
            RectTransform rt = canvasObj.GetComponent<RectTransform>();
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, Screen.width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, Screen.height);
            canvasObj.transform.localScale = new Vector3(0.0006f, 0.0006f, 0.0006f);
            canvasObj.transform.localPosition = new Vector3(0.0f, 0.0f, .5f);

            if (isUICanInteractive)
            {
                canvasObj.AddComponent<GraphicRaycaster>();
                SuperController.singleton.AddCanvas(popupCanvas);
            }

            rt.SetParent(headCenter, false);

            return popupCanvas;
        }

        public static Transform GetAtomRescaleObject(Atom atom)
        {
            return atom.containingAtom.reParentObject.Find("object/rescaleObject");
        }

        public static Vector2 MapScreenPosTo01(Vector3 screenPosition)
        {
            return new Vector2(screenPosition.x / (float)Screen.currentResolution.width, screenPosition.y / (float)Screen.currentResolution.height);
        }

        public static float GetVectorAngleRange360(Vector3 from, Vector3 to)
        {
            // Determine whether the vector of the cross product points off the screen to determine whether the rotation angle is greater than 180.
            Vector3 crossVec = Vector3.Cross(from, to);
            float dotVal = Vector3.Dot(crossVec, Vector3.back);
            float angle = Vector3.Angle(from, to);

            if (dotVal < 0)
            {
                angle = 360 - angle;
            }
            return angle;
        }

        // The projection of src in the direction of target
        public static Vector3 VecProject(Vector3 src, Vector3 target)
        {
            return Vector3.Dot(src, target) * target.normalized;
        }

        // Point of intersection of X-ray and plane
        public static Vector3 GetIntersectPointWithRay(Vector3 planePoint, Vector3 n, Ray ray)
        {
            float d = Vector3.Dot(planePoint, n);
            Vector3 rayStartPoint = ray.origin;
            Vector3 rayDirction = ray.direction.normalized;
            float t0 = (d - Vector3.Dot(rayStartPoint, n)) / Vector3.Dot(rayDirction, n);
            return rayStartPoint + rayDirction * t0;
        }

        // Get the vertical vector of the two-dimensional vector (point multiplied by 0, there are two)
        public static Vector2 GetVerticalVectorFrom(Vector2 from)
        {
            return new Vector2(-from.y, from.x).normalized;
        }

        const float minStartRadius = 0.001f;

        static int currentMeshSides;

        // Gets a procedural cone mesh with a specific number of sides.
        public static Mesh GetConeMesh(int desiredSides)
        {
            if (mesh == null || currentMeshSides != desiredSides)
            {
                DestroyOldMesh();

                // Generate a new cone mesh.
                mesh = GenerateCone(
                    height: 1f,
                    baseRadius: 1f,
                    topRadius: 1f,
                    sides: desiredSides,
                    segments: 35
                );

                currentMeshSides = desiredSides;
            }

            return mesh;
        }

        // Cone generation of complete rewriting
        public static Mesh GenerateCone(float height, float baseRadius, float topRadius, int sides, int segments)
        {
            var mesh = new Mesh();

            // Calculate the number of vertices on the side (+1 is because there are repeated first vertices at each end on the top and bottom circles to close the circle)
            int sideVertexCount = (segments + 1) * (sides + 1);

            // Calculate the number of triangles on one side.
            int sideTriangleCount = segments * sides * 2;

            // Prepare vertex, UV and triangle arrays
            Vector3[] vertices = new Vector3[sideVertexCount];
            Vector2[] uvs = new Vector2[sideVertexCount];
            int[] triangles = new int[sideTriangleCount * 3];

            for (int segment = 0; segment <= segments; segment++)
            {
                float segmentRatio = (float)segment / segments;
                float radius = Mathf.Lerp(baseRadius, topRadius, segmentRatio);
                float z = height * segmentRatio;

                for (int side = 0; side <= sides; side++)
                {
                    float angle = 2 * Mathf.PI * side / sides;
                    float x = Mathf.Cos(angle) * radius;
                    float y = Mathf.Sin(angle) * radius;
                    int index = segment * (sides + 1) + side;

                    vertices[index] = new Vector3(x, y, z);
                    uvs[index] = new Vector2(0, segmentRatio);
                }
            }


            int triangleIndex = 0;
            for (int segment = 0; segment < segments; segment++)
            {
                for (int side = 0; side < sides; side++)
                {
                    int current = segment * (sides + 1) + side;
                    int next = current + sides + 1;

                    // Add the first triangle
                    triangles[triangleIndex++] = current;
                    triangles[triangleIndex++] = current + 1;
                    triangles[triangleIndex++] = next + 1;

                    // Add a second triangle
                    triangles[triangleIndex++] = current;
                    triangles[triangleIndex++] = next + 1;
                    triangles[triangleIndex++] = next;
                }
            }

            mesh.vertices = vertices;
            mesh.uv = uvs;
            mesh.triangles = triangles;
            mesh.RecalculateNormals();
            mesh.RecalculateBounds();

            return mesh;
        }

        public static void DestroyOldMesh()
        {
            if (mesh != null)
            {
                GameObject.DestroyImmediate(mesh);
                mesh = null;
            }
        }

        static Mesh mesh = null;

        public static T GetSafeComponent<T>(GameObject obj) where T : Component
        {
            T comp = obj.GetComponent<T>();
            if (comp == null)
            {
                comp = obj.AddComponent<T>();
            }
            return comp;
        }

		// Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
		public static string GetPluginPath(MVRScript self)
		{
			string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}

        public static JSONStorableBool CreateToggle(MVRScript self, string text, bool startValue, bool rightSide, bool acs)
        {
            JSONStorableBool storable = new JSONStorableBool(text, startValue);
            storable.storeType = JSONStorableParam.StoreType.Full;
            self.CreateToggle(storable, rightSide).label = Location.GetText(text);
            self.RegisterBool(storable);


            Toggle toggle = storable.toggle;
            toggle.enabled = acs;

            if (!acs)
                toggle.transform.Find("Panel").GetComponent<Image>().color = Color.gray;

            return storable;
        }

        // Create VaM-UI Float slider
        public static JSONStorableFloat CreateSliderFloat(MVRScript self, string text, float startValue, float min, float max, bool rightSide, bool acs)
        {
            JSONStorableFloat storable = new JSONStorableFloat(text, startValue, min, max, true, true);
            storable.storeType = JSONStorableParam.StoreType.Full;
            self.CreateSlider(storable, rightSide).label = Location.GetText(text);
            self.RegisterFloat(storable);

            Slider slider = storable.slider;
            slider.enabled = acs;

            if (!acs)
                slider.transform.parent.Find("Panel").GetComponent<Image>().color = Color.gray;

            return storable;
        }

        // Create VaM-UI Float slider
        public static JSONStorableFloat CreateSliderInt(MVRScript self, string text, int startValue, int min, int max, bool rightSide, bool acs)
        {
            JSONStorableFloat storable = new JSONStorableFloat(text, startValue, min, max, true, true);
            storable.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider slider = self.CreateSlider(storable, rightSide);
            slider.label = Location.GetText(text);
            slider.slider.wholeNumbers = true;
            slider.valueFormat = "F0";
            self.RegisterFloat(storable);

            Slider slider1 = storable.slider;
            slider1.enabled = acs;

            if (!acs)
                slider1.transform.parent.Find("Panel").GetComponent<Image>().color = Color.gray;

            return storable;
        }

        // Create VaM-UI ColorPicker
        public static JSONStorableColor CreateColorPicker(MVRScript script, string label, Color color, bool rightSide)
        {
            HSVColor hsvColor = HSVColorPicker.RGBToHSV(color.r, color.g, color.b);
            JSONStorableColor storable = new JSONStorableColor(label, hsvColor);
            storable.storeType = JSONStorableParam.StoreType.Full;
            script.CreateColorPicker(storable, rightSide).label = Location.GetText(label);
            script.RegisterColor(storable);
            return storable;
        }

        // Create VaM-UI StringChooser
        public static JSONStorableStringChooser CreateStringChooser(MVRScript self, string text, List<string> entries, bool rightSide, bool acs)
        {
            string defaultEntry = entries.Count > 0 ? entries[0] : "";
            JSONStorableStringChooser storable = new JSONStorableStringChooser(text, entries, defaultEntry, text);
            var popUI = self.CreateScrollablePopup(storable, rightSide);
            popUI.label = Location.GetText(text);
            self.RegisterStringChooser(storable);

            Image image = popUI.transform.Find("Button").GetComponentInChildren<Image>();
            image.raycastTarget = acs;

            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        // Create VaM-UI StringChooser
        public static JSONStorableStringChooser CreateStringChooser(MVRScript self, string text, List<string> entries, int defaultIndex, bool rightSide, bool acs)
        {
            string defaultEntry = (defaultIndex >= 0 && defaultIndex < entries.Count) ? entries[defaultIndex] : "";
            JSONStorableStringChooser storable = new JSONStorableStringChooser(text, entries, defaultEntry, text);
            var popUI = self.CreateScrollablePopup(storable, rightSide);
            popUI.label = Location.GetText(text);
            self.RegisterStringChooser(storable);

            Image image = popUI.transform.Find("Button").GetComponentInChildren<Image>();
            image.raycastTarget = acs;

            if (!acs)
                image.color = Color.gray;

            return storable;
        }

        // Create VaM-UI InfoText field
        public static JSONStorableString CreateInfoText(MVRScript self, string text, float height, bool rightSide)
        {
            JSONStorableString storable = new JSONStorableString("Info", text);
            UIDynamicTextField textfield = self.CreateTextField(storable, rightSide);
            textfield.text = Location.GetText(text);
            textfield.height = height;
            return storable;
        }

        public static UIDynamic CreateSpacer(MVRScript script, float height, bool rightSide)
        {
            UIDynamic spacer = script.CreateSpacer(rightSide);
            spacer.height = height;
            return spacer;
        }

        // Create VaM-UI button
        public static UIDynamicButton CreateButton(MVRScript self, string text, UnityAction callback, bool rightSide, bool acs)
        {
            UIDynamicButton btn = self.CreateButton(text, rightSide);
            btn.button.onClick.AddListener(callback);
            btn.buttonText.text = Location.GetText(text);

            btn.button.enabled = acs;

            if (!acs)
            {
                Image image = btn.button.GetComponent<Image>();
                image.color = Color.gray;
            }

            return btn;
        }

        // Create input action trigger
        public static JSONStorableAction SetupAction(MVRScript self, string name, JSONStorableAction.ActionCallback callback)
        {
            JSONStorableAction action = new JSONStorableAction(name, callback);
            self.RegisterAction(action);
            return action;
        }

        // ===========================================================================================
        // Custom UI system with new UI elements and the ability to easily add/remove UI at runtime
        //
        // Usage instructions:
        // - Before using the custom UI elements, call from your MVRScript:
        //       Utils.OnInitUI(CreateUIElement);
        // - When your MVRScript receives the OnDestroy message call:
        //       Utils.OnDestroyUI();

        // Create one-line text input with label
        public static UIDynamicLabelInput SetupTextInput(MVRScript self, string text, JSONStorableString storable, bool rightSide)
        {
			if (internalInputPrefab == null)
            {
                CreateInternalInputPrefab(self);
            }

            Transform inputTrans = internalCreateUIElement(internalInputPrefab.transform, rightSide);
            UIDynamicLabelInput dinput = inputTrans.gameObject.GetComponent<UIDynamicLabelInput>();
            storable.inputField = dinput.input;
            dinput.text.text = Location.GetText(text);
            inputTrans.gameObject.SetActive(true);
            return dinput;
        }

        private static void CreateInternalInputPrefab(MVRScript self)
        {
            internalInputPrefab = new GameObject("LabelInput");
            internalInputPrefab.SetActive(false);
            RectTransform rt = internalInputPrefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            rt.offsetMax = new Vector2(535, -500);
            rt.offsetMin = new Vector2(10, -600);
            LayoutElement le = internalInputPrefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = 45;
            le.minWidth = 350;
            le.preferredHeight = 45;
            le.preferredWidth = 500;

            RectTransform backgroundTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform = UnityEngine.Object.Instantiate(backgroundTransform, internalInputPrefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            backgroundTransform.offsetMax = new Vector2(0, 0);
            backgroundTransform.offsetMin = new Vector2(0, -10);

            RectTransform labelTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Button/Text") as RectTransform; ;
            labelTransform = UnityEngine.Object.Instantiate(labelTransform, internalInputPrefab.transform);
            labelTransform.name = "Text";
            labelTransform.anchorMax = new Vector2(0, 1);
            labelTransform.anchorMin = new Vector2(0, 0);
            labelTransform.offsetMax = new Vector2(155, -5);
            labelTransform.offsetMin = new Vector2(5, 0);
            Text labelText = labelTransform.GetComponent<Text>();
            labelText.text = "Name";
            labelText.color = Color.white;

            RectTransform inputTransform = self.manager.configurableTextFieldPrefab.transform as RectTransform;
            inputTransform = UnityEngine.Object.Instantiate(inputTransform, internalInputPrefab.transform);
            inputTransform.anchorMax = new Vector2(1, 1);
            inputTransform.anchorMin = new Vector2(0, 0);
            inputTransform.offsetMax = new Vector2(-5, -5);
            inputTransform.offsetMin = new Vector2(160, -5);
            UIDynamicTextField textfield = inputTransform.GetComponent<UIDynamicTextField>();
            textfield.backgroundColor = Color.white;
            LayoutElement layout = textfield.GetComponent<LayoutElement>();
            layout.preferredHeight = layout.minHeight = 35;
            InputField inputfield = textfield.gameObject.AddComponent<InputField>();
            inputfield.textComponent = textfield.UItext;
            inputfield.textComponent.resizeTextForBestFit = true;

            RectTransform textTransform = textfield.UItext.rectTransform;
            textTransform.anchorMax = new Vector2(1, 1);
            textTransform.anchorMin = new Vector2(0, 0);
            textTransform.offsetMax = new Vector2(-5, -5);
            textTransform.offsetMin = new Vector2(10, -5);

            UnityEngine.Object.Destroy(textfield);

            UIDynamicLabelInput uid = internalInputPrefab.AddComponent<UIDynamicLabelInput>();
            uid.text = labelText;
            uid.input = inputfield;
        }

        public static UIDynamicTextInfo CreateInfoTextNoScroll(MVRScript self, string text, float height, bool rightSide)
        {
			if (internalTextInfoPrefab == null)
            {
                CreateInternalTextInfoPrefab(self);
            }

            Transform textTrans = internalCreateUIElement(internalTextInfoPrefab.transform, rightSide);
            UIDynamicTextInfo dtext = textTrans.gameObject.GetComponent<UIDynamicTextInfo>();
            dtext.text.text = Location.GetText(text);
            dtext.layout.minHeight = height;
            dtext.layout.preferredHeight = height;
            textTrans.gameObject.SetActive(true);
            return dtext;
        }

        private static void CreateInternalTextInfoPrefab(MVRScript self)
        {
            internalTextInfoPrefab = new GameObject("TextInfo");
            internalTextInfoPrefab.SetActive(false);
            RectTransform rt = internalTextInfoPrefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            rt.offsetMax = new Vector2(535, -500);
            rt.offsetMin = new Vector2(10, -600);
            LayoutElement le = internalTextInfoPrefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = 35;
            le.minWidth = 350;
            le.preferredHeight = 35;
            le.preferredWidth = 500;

            RectTransform backgroundTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform = UnityEngine.Object.Instantiate(backgroundTransform, internalTextInfoPrefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            backgroundTransform.offsetMax = new Vector2(0, 0);
            backgroundTransform.offsetMin = new Vector2(0, -10);

            RectTransform labelTransform = self.manager.configurableScrollablePopupPrefab.transform.Find("Button/Text") as RectTransform; ;
            labelTransform = UnityEngine.Object.Instantiate(labelTransform, internalTextInfoPrefab.transform);
            labelTransform.name = "Text";
            labelTransform.anchorMax = new Vector2(1, 1);
            labelTransform.anchorMin = new Vector2(0, 0);
            labelTransform.offsetMax = new Vector2(-5, 0);
            labelTransform.offsetMin = new Vector2(5, 0);
            Text labelText = labelTransform.GetComponent<Text>();
            labelText.alignment = TextAnchor.UpperLeft;

            UIDynamicTextInfo uid = internalTextInfoPrefab.AddComponent<UIDynamicTextInfo>();
            uid.text = labelText;
            uid.layout = le;
            uid.background = backgroundTransform;
        }

        public static UIDynamicTextInfo CreateInfoTextNoScroll(MVRScript self, JSONStorableString storable, float height, bool rightSide)
		{
			UIDynamicTextInfo uid = CreateInfoTextNoScroll(self, storable.val, height, rightSide);
			storable.setCallbackFunction = (string text) => {
                if (uid != null && uid.text != null)
                {
                    uid.text.text = Location.GetText(text);
                }
			};
			return uid;
		}

		public delegate Transform CreateUIElement(Transform prefab, bool rightSide);
		public static void InitUI(CreateUIElement createUIElementCallback)
		{
			internalCreateUIElement = createUIElementCallback;
		}

		public static void DestroyUI()
		{
			SafeDestroy(ref internalInputPrefab);
			SafeDestroy(ref internalTextInfoPrefab);
		}

		private static void SafeDestroy(ref GameObject go)
		{
			if (go != null) GameObject.Destroy(go);
        }

		private static CreateUIElement internalCreateUIElement;
		private static GameObject internalInputPrefab;
		private static GameObject internalTextInfoPrefab;
	}

	// ===========================================================================================

	public delegate void TextureSetCallback(Texture2D tex);

	public class TextureSettings
	{
		public bool compress = false;
		public bool createMipMaps = true;
		public bool isNormalMap = false;
		public bool linearColor = true; // Using linear or sRGB color space.
		public bool createAlphaFromGrayscale = false;
		public bool createNormalFromBump = false;
		public float bumpStrength = 1.0f;
		public TextureWrapMode wrapMode = TextureWrapMode.Repeat;
		public FilterMode filterMode = FilterMode.Trilinear;
		public int anisoLevel = 5; // 0: Forced off, 1: Off, quality setting can override, 2-9: Anisotropic filtering levels.
	}

    public class UIDynamicBase : UIDynamic
    {
    }

    public class UIDynamicLabelInput : UIDynamicBase
    {
        public Text text;
        public InputField input;
    }

    public class UIDynamicTextInfo : UIDynamicBase
    {
        public Text text;
        public LayoutElement layout;
        public RectTransform background;
    }

    public struct PlaneAlignment
    {
        /// <summary>Align the plane to the surface normal which blocks the beam. Works better for large occluders such as floors and walls.</summary>
        public const int Surface = 0;
        /// <summary>Keep the plane aligned with the beam direction. Workds better with more complex occluders or with corners.</summary>
        public const int Beam = 1;
    }

    public struct ColorMode
    {
        // Apply a flat/plain/single color
        public const int Flat = 0;
        // Apply a gradient
        public const int Gradient = 1;
    }

    public struct AttenuationEquation
    {
        // Simple linear attenuation.
        public const int Linear = 0;
        // Quadratic attenuation, which usually gives more realistic results.
        public const int Quadratic = 1;
        // Custom blending mix between linear and quadratic attenuation formulas. Use attenuationEquation property to tweak the mix.
        public const int Blend = 3;
    }

    public struct BlendingMode
    {
        public const int Additive = 0;
        public const int SoftAdditive = 1;
        public const int TraditionalTransparency = 2;
    }

    public struct MeshType
    {
        // Use the global shared mesh (recommended setting, since it will save a lot on memory). Will use the geometry properties set on Config.
        public const int Shared = 0;
        // Use a custom mesh instead. Will use the geometry properties set on the beam.
        public const int Custom = 1;
    }

    public struct RenderingMode
    {
        /// <summary> Use the 2 pass shader. Will generate 2 drawcalls per beam (Not compatible with custom Render Pipeline such as HDRP and LWRP).</summary>
        public const int MultiPass = 0;
        /// <summary> Use the 1 pass shader. Will generate 1 drawcall per beam. </summary>
        public const int SinglePass = 1;
        /// <summary> Dynamically batch multiple beams to combine and reduce draw calls. </summary>
        public const int GPUInstancing = 2;
    }

    public struct RenderQueue
    {
        /// Specify a custom render queue.
        public const int Custom = 0;

        /// This render queue is rendered before any others.
        public const int Background = 1000;

        /// Opaque geometry uses this queue.
        public const int Geometry = 2000;

        /// Alpha tested geometry uses this queue.
        public const int AlphaTest = 2450;

        /// Last render queue that is considered "opaque".
        public const int GeometryLast = 2500;

        /// This render queue is rendered after Geometry and AlphaTest, in back-to-front order.
        public const int Transparent = 3000;

        /// This render queue is meant for overlay effects.
        public const int Overlay = 4000;
    }

    public struct OccluderDimensions
    {
        /// <summary> the beam will react against 3D Occluders. </summary>
        public const int Occluders3D = 0;

        /// <summary> the beam will react against 2D Occluders. This is useful when using the beams with 2D objects (such as 2D Sprites). </summary>
        public const int Occluders2D = 1;
    }
}
