﻿using UnityEngine;
using System.Collections.Generic;
using Battlehub.RTCommon;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using System;
using UnityEngine.EventSystems;
using UnityEngine.UI;
using System.Collections;
using Object = UnityEngine.Object;
using MVR.FileManagementSecure;
using System.Security.Cryptography;
using System.Text;

namespace PluginIdea.VolumeLight
{

    public struct AttachType
    {
        public const int Light = 0;
        public const int Empty = 1;
        public const int Parent = 2;
    }

    public class LightFXInfo
    {
        public Atom atom;
        public Transform lightTrans;
        public Renderer ballRenderer;
    }

    class HPVLB : MVRScript
    {

        public static Dictionary<string, Shader> dic_name2shader = new Dictionary<string, Shader>();
        public static Dictionary<string, GameObject> dic_name2GameObject = new Dictionary<string, GameObject>();
        public static Dictionary<string, Texture3D> dic_name2Object = new Dictionary<string, Texture3D>();

        private const string Key_Active = "cer";
        private JSONStorableString aci;
        private string info;

        private JSONStorableBool isEnableLightSphere;
        private JSONStorableFloat extraIntensityWhenDazzle;
        private JSONStorableFloat lightSphereBaseIntensity;
        private JSONStorableFloat lightSphereRadiusMultiper;
        private JSONStorableFloat lightSphereGlowPower;
        private JSONStorableFloat lightSphereIntensityMultiper;
        private JSONStorableFloat lightSphereOffsetZ;
        private List<GameObject> lightSphereParents = new List<GameObject>();

        private bool isFree = true;

        private Request request;
        private bool isResourcesLoaded;

        public bool acs;
        public override void Init()
        {
            Location.InitTData();
            InitLightTrans();
            UIInit();
            DataInit();
            RegistBundleUnload();
        }

        private Button hardResetBtn;
        private void RegistBundleUnload()
        {
            hardResetBtn = SuperController.singleton.mainHUD.Find("MainUICanvas/Panel/Content/TabFile/Content/Hard Reset Button").GetComponent<Button>();
            hardResetBtn.onClick.AddListener(UnloadBundle);
        }

        private void UnloadBundle()
        {
            if (request != null && request.assetBundle != null)
            {
                Debug.Log("Unload ---------------------");
                request.assetBundle.Unload(false);
            }
        }

        private void UIInit()
        {
            Utils.InitUI(CreateUIElement);

            //Utils.CreateSpacer(this, 50, true);
            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.PluginInfo), 200, true);

            // ----------------------------------------------------------------------------------
            aci = new JSONStorableString("ac", Location.GetText(Location.GetStr(Location.ACInfo)));

            acs = false;

            acs = isFree;

            if (isFree)
            {
                var btn = Utils.CreateButton(this, Location.GetStr(Location.OpenMainPage), () => {
                    if (Location.IsChinease) Application.OpenURL(@"https:\/\/afdian.net/p/6f6685ceae9211ed8c8952540025c377");
                    else Application.OpenURL(@"https:\/\/www.patreon.com/PluginIdea/about");
                }, false, true);
                btn.buttonImage.color = Color.yellow;
                Utils.CreateSpacer(this, 30, false);
            }

            var aff = acs;
            // ----------------------------------------------------------------------------------

            Utils.CreateSpacer(this, 20, true);
            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.LightSphereSubtitle), 40, true);

            isEnableLightSphere = Utils.CreateToggle(this, Location.GetStr(Location.EnableLightSphere), true, true, acs);
            isEnableLightSphere.setCallbackFunction = (val) => {
                if (lightTranses.Count > 0) lightTranses.ForEach(lt => { if(lt != null && lt.ballRenderer != null) lt.ballRenderer.gameObject.SetActive(val);
                });
            };

            extraIntensityWhenDazzle = Utils.CreateSliderFloat(this, Location.GetStr(Location.ExtraWhenDazzle),0.2f, 0f, 1.5f, true, acs);
            lightSphereBaseIntensity = Utils.CreateSliderFloat(this, Location.GetStr(Location.LightSphereBaseIntensity), 0.2f, 0f, 1, true, acs);
            lightSphereRadiusMultiper = Utils.CreateSliderFloat(this, Location.GetStr(Location.LightSphereRadiusMultiper), 2f, 1f, 3, true, acs);
            lightSphereGlowPower = Utils.CreateSliderFloat(this, Location.GetStr(Location.LightGlowPower), 2.5f, 0.5f, 4f, true, acs);
            lightSphereIntensityMultiper = Utils.CreateSliderFloat(this, Location.GetStr(Location.LightSphereIntensityMultiper), 3f, 1,6f, true, acs);
            lightSphereGlowPower.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (lightTranses.Count > 0) lightTranses.ForEach(lt => { if (lt != null && lt.ballRenderer != null) lt.ballRenderer.material.SetFloat("_MKGlowPower", val); });
            };

            lightSphereOffsetZ = Utils.CreateSliderFloat(this, Location.GetStr(Location.LightSphereOffsetZ), -0.02f, -0.5f, 0.5f, true, acs);
            lightSphereOffsetZ.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (lightTranses.Count > 0) lightTranses.ForEach(lt => { if (lt != null && lt.ballRenderer != null) lt.ballRenderer.transform.localPosition = new Vector3(0, 0, val); });
            };

            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.LightBeamSubtitle), 40, false);
            isEnableLightBeam = Utils.CreateToggle(this, Location.GetStr(Location.IsEnableLightBeam), true, false, acs);
            isEnableLightBeam.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.IsEnableLight = val; });
            };

            isIndependentControl = Utils.CreateToggle(this, Location.GetStr(Location.IsIndependentControl), false, false, acs);

            isRealtime = Utils.CreateToggle(this, Location.GetStr(Location.IsRealTime), true, false, acs);
            isRealtime.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.isUpdatePropertyFromLight = val; });
            };

            Utils.CreateSpacer(this, 20, true);
            if (attachType == AttachType.Light)
            {
                colorFromLight = Utils.CreateToggle(this, Location.GetStr(Location.ColorFromLight), true, true, aff);
                colorFromLight.setCallbackFunction = (val) => {
                    if (!isResourcesLoaded) return;
                    if (vbs.Count > 0) vbs.ForEach(vb => { vb.useLightColor = val; });
                };
            }

            color = Utils.CreateColorPicker(this, Location.GetStr(Location.CustomColor), Color.white, true);
            color.setCallbackFunction = (h,s,v) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.mainColor = Color.HSVToRGB(h, s, v); });
            };

            Utils.CreateSpacer(this, 20, false);
            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.AlphaSubtitle), 35, false);
            if (attachType == AttachType.Light)
            {
                alphaInOutFromLight = Utils.CreateToggle(this, Location.GetStr(Location.AlphaInOutFromLight), false, false, acs);
                alphaInOutFromLight.setCallbackFunction = (val) => {
                    if (!isResourcesLoaded) return;
                    if (vbs.Count > 0) vbs.ForEach(vb => { vb.deriveAlphaFromIntensity = val; });
                };
            }

            intensityMultiplier = Utils.CreateSliderFloat(this, Location.GetStr(Location.IntensityMultiper), 1f, 0f, 10, false, acs);
            intensityMultiplier.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.intensityFactor = val; });
            };
            if (attachType == AttachType.Light)
            {
                alphamultiplier = Utils.CreateSliderFloat(this, Location.GetStr(Location.AlphaMultiplier), 1.5f, 0.1f, 3, false, acs);
                alphamultiplier.setCallbackFunction = (val) => {
                    if (!isResourcesLoaded) return;
                    if (vbs.Count > 0) vbs.ForEach(vb => { vb.alphaScale = val; });
                };
            }

            alphaInside = Utils.CreateSliderFloat(this, Location.GetStr(Location.AlphaInside), 1, 0, 1, false, acs);
            alphaInside.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.internalAlpha = val; });
            };

            alphaOutside = Utils.CreateSliderFloat(this, Location.GetStr(Location.AlphaOutside), 1, 0, 1, false, acs);
            alphaOutside.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.externalAlpha = val; });
            };

            Utils.CreateSpacer(this, 20, false);
            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.SpotAngleSubleTitle), 35, false);
            if (attachType == AttachType.Light)
            {
                spotAngleFromLight = Utils.CreateToggle(this, Location.GetStr(Location.SpotAngleFromLight), true, false, acs);
                spotAngleFromLight.setCallbackFunction = (val) => {
                    if (!isResourcesLoaded) return;
                    if (vbs.Count > 0) vbs.ForEach(vb => { vb.deriveAngleFromLight = val; });
                };
            }

            spotAngle = Utils.CreateSliderFloat(this, Location.GetStr(Location.CustomSpotAngle), 50, 0, 180, false, acs);
            spotAngle.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.spotlightAngle = val; });
            };

            Utils.CreateSpacer(this, 20, false);
            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.FadeSubtitle), 35, false);
            if (attachType == AttachType.Light)
            {
                fadeFromLight = Utils.CreateToggle(this, Location.GetStr(Location.FadeFromLight), true, false, acs);
                fadeFromLight.setCallbackFunction = (val) => {
                    if (!isResourcesLoaded) return;
                    if (vbs.Count > 0) vbs.ForEach(vb => { vb.deriveFadeEndFromLight = val; });
                };
            }

            fadeStart = Utils.CreateSliderFloat(this, Location.GetStr(Location.CustomFadeStart), 0, 0, 30, false, acs);
            fadeStart.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.lightFadeStart = val; });
            };

            fadeEnd = Utils.CreateSliderFloat(this, Location.GetStr(Location.CustomFadeEnd), 10, 0, 60, false, acs);
            fadeEnd.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.lightFadeEnd = val; });
            };

            Utils.CreateSpacer(this, 20, false);
            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.ConeShapeSubtitle), 35, false);
            coneScaleY = Utils.CreateSliderFloat(this, Location.GetStr(Location.ConeScaleY), 1, 0.001f, 1, false, acs);
            coneScaleY.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.transform.localScale = new Vector3(vb.transform.localScale.x, val, vb.transform.localScale.z); });
            };

            coneRadiusStart = Utils.CreateSliderFloat(this, Location.GetStr(Location.ConeRadiusStart), 0.05f, 0, 1, false, acs);
            coneRadiusStart.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (lightTranses.Count > 0) lightTranses.ForEach(lt => { if (lt != null && lt.ballRenderer != null) lt.ballRenderer.transform.localScale = Vector3.one * val * lightSphereRadiusMultiper.val; });
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.startConeRadius = val; });
            };

            sidesAmount = Utils.CreateSliderInt(this, Location.GetStr(Location.ConeSegAmount), 40, 3, 50, false, acs);
            sidesAmount.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.numOfSides = (int)val; vb.CreateGeometry(); });
            };

            Utils.CreateInfoTextNoScroll(this, Location.GetStr(Location.NoiseSubtitle), 40, true);

            noiseIntensity = Utils.CreateSliderFloat(this, Location.GetStr(Location.NoiseIntensity), 1f, 0f, 1f, true, acs);
            noiseIntensity.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.noiseStrength = val; });
            };

            noiseScale = Utils.CreateSliderFloat(this, Location.GetStr(Location.NoiseScale), 0.61f, 0.01f, 2, true, acs);
            noiseScale.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.noiseScale = val; });
            };

            noiseSpeedX = Utils.CreateSliderFloat(this, Location.GetStr(Location.NoiseSpeedX), 0.1f, 0f, 0.2f, true, acs);
            noiseSpeedX.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.noiseSpeed = new Vector3(val, vb.noiseSpeed.y, vb.noiseSpeed.z); });
            };

            noiseSpeedY = Utils.CreateSliderFloat(this, Location.GetStr(Location.NoiseSpeedY), 0, 0f, 0.2f, true, acs);
            noiseSpeedY.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.noiseSpeed = new Vector3(vb.noiseSpeed.x, val, vb.noiseSpeed.z); });
            };

            noiseSpeedZ = Utils.CreateSliderFloat(this, Location.GetStr(Location.NoiseSpeedZ), 0, 0f, 0.2f, true, acs);
            noiseSpeedZ.setCallbackFunction = (val) => {
                if (!isResourcesLoaded) return;
                if (vbs.Count > 0) vbs.ForEach(vb => { vb.noiseSpeed = new Vector3(vb.noiseSpeed.x, vb.noiseSpeed.y, val); });
            };
        }

        private int attachType = AttachType.Light;
        private List<LightFXInfo> lightTranses = new List<LightFXInfo>();
        private void DataInit()
        {
            string assetsBundlePath = Utils.GetPluginPath(this) + string.Format("/resources/hpvl{0}.assetbundle", Utils.version);
            request = new Request { path = assetsBundlePath, callback = OnAssetBundleLoaded };
            AssetLoader.QueueLoadAssetBundleFromFile(request);
        }

        private void InitLightTrans()
        {
            lightTranses.Clear();
            List<Atom> childAtoms = new List<Atom>();
            Utils.GetAllEmptyLightChildsFromAtom(containingAtom, childAtoms);
            if (childAtoms.Count > 0)
            {
                attachType = AttachType.Parent;
                childAtoms.ForEach(at => {
                    if (at.type == "Empty")
                    {
                        CreateEmptyLightTrans(at);
                    }
                    //else if (containingAtom.reParentObject.GetComponentInChildren<Light>(true) != null)
                    //{
                    //    AppendLightComp(at);
                    //}
                });
            }
            else if (containingAtom.type == "Empty")
            {
                attachType = AttachType.Empty;
                CreateEmptyLightTrans(containingAtom);
            }
            else if (containingAtom.reParentObject.GetComponentInChildren<Light>(true) != null)
            {
                attachType = AttachType.Light;
                AppendLightComp(containingAtom);
            }
        }

        private void AppendLightComp(Atom atom)
        {
            Light light = atom.reParentObject.GetComponentInChildren<Light>(true);
            lightTranses.Add(new LightFXInfo() { lightTrans = light.transform , atom = atom});
        }

        private Transform CreateEmptyLightTrans(Atom emptyAtom)
        {
            Transform lightTrans = new GameObject("VLLightTrans").transform;
            lightTranses.Add(new LightFXInfo() { lightTrans = lightTrans, atom = emptyAtom });
            Transform rescaleTrans = emptyAtom.reParentObject.Find("object/rescaleObject");
            rescaleTrans.Find("SphereRender").GetComponent<Renderer>().enabled = false;
            lightTrans.SetParent(rescaleTrans);
            lightTrans.localPosition = Vector3.zero;
            lightTrans.localRotation = Quaternion.identity;
            lightTrans.localScale = Vector3.one;
            return lightTrans;
        }

        private void OnAssetBundleLoaded(Request abffr)
        {

            Shader[] shaders = abffr.assetBundle.LoadAllAssets<Shader>();
            if (shaders.Length > 0)
            {
                for (int i = 0; i < shaders.Length; i++)
                {
                    if (!dic_name2shader.ContainsKey(shaders[i].name)) dic_name2shader.Add(shaders[i].name, shaders[i]);
                }
            }

            GameObject[] gobjs = abffr.assetBundle.LoadAllAssets<GameObject>();
            gobjs.ToList().ForEach(x => {
                if (!dic_name2GameObject.ContainsKey(x.name)) dic_name2GameObject.Add(x.name, x);
            });

            Texture3D[] objs = abffr.assetBundle.LoadAllAssets<Texture3D>();
            objs.ToList().ForEach(x => {
                if (!dic_name2Object.ContainsKey(x.name)) dic_name2Object.Add(x.name, x);
            });

            isResourcesLoaded = true;
            ReApplyUIParameter();
        }

        List<VolumetricLightMono> vbs = new List<VolumetricLightMono>();
        private JSONStorableFloat alphamultiplier;
        private JSONStorableFloat intensityMultiplier;
        private JSONStorableFloat alphaInside;
        private JSONStorableFloat alphaOutside;
        private JSONStorableBool isRealtime;
        private JSONStorableBool isEnableLightBeam;
        private JSONStorableFloat noiseIntensity;

        private JSONStorableBool isIndependentControl;

        private JSONStorableBool colorFromLight;
        private JSONStorableBool alphaInOutFromLight;
        private JSONStorableColor color;
        private JSONStorableBool spotAngleFromLight;
        private JSONStorableFloat spotAngle;
        private JSONStorableBool fadeFromLight;
        private JSONStorableFloat fadeStart;
        private JSONStorableFloat fadeEnd;
        private JSONStorableFloat coneRadiusStart;
        private JSONStorableFloat coneScaleY;
        private JSONStorableFloat sidesAmount;
        private JSONStorableFloat segmentsAmount;
        private JSONStorableFloat noiseScale;
        private JSONStorableFloat noiseSpeedX, noiseSpeedY, noiseSpeedZ;

        private void ReApplyUIParameter()
        {
            if (lightTranses.Count > 0)
            {
                lightTranses.ForEach(lt => {
                    vbs.Add(Utils.GetSafeComponent<VolumetricLightMono>(lt.lightTrans.gameObject));

                    GameObject lightSphereParent = new GameObject("LightObjParent");
                    lightSphereParents.Add(lightSphereParent);
                    GameObject lightSphere = GameObject.CreatePrimitive(PrimitiveType.Sphere);
                    lightSphere.transform.SetParent(lightSphereParent.transform);
                    lightSphere.transform.localPosition = Vector3.zero;
                    lightSphere.transform.localEulerAngles = new Vector3(-90, 0, 0);
                    var lightSphereRender = lightSphere.GetComponent<Renderer>();
                    lightSphereRender.shadowCastingMode = UnityEngine.Rendering.ShadowCastingMode.Off;
                    lt.ballRenderer = lightSphereRender;
                    lightSphereParent.transform.SetParent(lt.lightTrans);
                    lightSphereParent.transform.localPosition = Vector3.zero;
                    lightSphereParent.transform.localRotation = Quaternion.identity;

                    lightSphereRender.material.shader = Shader.Find("Custom/Subsurface/EmissiveGlow");
                    Utils.GetSafeComponent<MKGlowObject>(lightSphereRender.gameObject);
                });
            }
            else
            {
                SuperController.LogError("Can't Find Empty/Light Atom!");
            }
            sidesAmount.setCallbackFunction(sidesAmount.val);
            isEnableLightSphere.setCallbackFunction(isEnableLightSphere.val);
            lightSphereOffsetZ.setCallbackFunction(lightSphereOffsetZ.val);
            lightSphereGlowPower.setCallbackFunction(lightSphereGlowPower.val);
          
            intensityMultiplier.setCallbackFunction(intensityMultiplier.val);
            alphaInside.setCallbackFunction(alphaInside.val);
            alphaOutside.setCallbackFunction(alphaOutside.val);
            color.setCallbackFunction(color.val.H, color.val.S, color.val.V);
           
            spotAngle.setCallbackFunction(spotAngle.val);

            coneScaleY.setCallbackFunction(coneScaleY.val);
            fadeStart.setCallbackFunction(fadeStart.val);
            fadeEnd.setCallbackFunction(fadeEnd.val);
            coneRadiusStart.setCallbackFunction(coneRadiusStart.val);
            noiseScale.setCallbackFunction(noiseScale.val);
            noiseIntensity.setCallbackFunction(noiseIntensity.val);

            noiseSpeedX.setCallbackFunction(noiseSpeedX.val);
            noiseSpeedY.setCallbackFunction(noiseSpeedY.val);
            noiseSpeedZ.setCallbackFunction(noiseSpeedZ.val);

            isRealtime.setCallbackFunction(isRealtime.val);

            if (attachType == AttachType.Light)
            {
                alphamultiplier.setCallbackFunction(alphamultiplier.val);
                alphaInOutFromLight.setCallbackFunction(alphaInOutFromLight.val);
                colorFromLight.setCallbackFunction(colorFromLight.val);
                spotAngleFromLight.setCallbackFunction(spotAngleFromLight.val);
                fadeFromLight.setCallbackFunction(fadeFromLight.val);
            }
        }

        private void Update()
        {
            if (isRealtime.val)
            {
                if (lightTranses != null && lightTranses.Count > 0)
                {
                    lightTranses.ForEach(lt => {
                        if (lt != null)
                        {
                            if (lt.lightTrans != null && lt.ballRenderer != null)
                            {
                                float angle = Vector3.Angle(-Utils.CenterCameraTrans.forward, lt.lightTrans.forward);
                                float dot = Vector3.Dot(-Utils.CenterCameraTrans.forward, lt.lightTrans.forward);

                                var dfrictionVal = lt.atom.GetStorableByID("PhysicsMaterialControl").GetFloatJSONParam("dynamicFriction").val;
                                //var sfrictionVal = lt.atom.GetStorableByID("PhysicsMaterialControl").GetFloatJSONParam("staticFriction").val;

                                Light light = lt.lightTrans.GetComponent<Light>();
                                float sa = light != null ? light.spotAngle : spotAngle.val;
                                float extraIntensity = dot > 0 && angle < sa * 0.5f ? extraIntensityWhenDazzle.val : 0;
                                float intensityResult = Mathf.Max(0, dot) + lightSphereBaseIntensity.val + extraIntensity;

                                float h, s, v;
                                if (light != null)
                                {
                                    Color.RGBToHSV(light.color, out h, out s, out v);
                                }
                                else
                                {
                                    h = color.val.H;
                                    s = color.val.S;
                                    v = color.val.V;
                                }

                                float intensity = light != null ? light.intensity : 1;

                                // 启动动摩擦参数来接受独立参数
                                intensity *= isIndependentControl.val ? dfrictionVal : 1;

                                v *= intensityResult * lightSphereIntensityMultiper.val * intensity;

                                Color fcolor = Color.HSVToRGB(h, s, v);
                                lt.ballRenderer.material.SetColor("_MKGlowColor", fcolor);
                                lt.ballRenderer.material.SetColor("_Color", fcolor);

                                var vb = lt.lightTrans.GetComponentInChildren<VolumetricLightMono>();
                                if (vb) vb.lightIntensity = intensity;
                            }
                        }
                    });
                }
            }
        }

        private string GMAA()
        {
            var gid = SystemInfo.graphicsDeviceID;
            var vf = SystemInfo.graphicsDeviceVendorID;
            var rdd = UnityEngine.Random.Range(10000, 999990).GetHashCode();
            return string.Format("{0}_{1}_{2}", gid, vf, rdd).ToString();
        }

        private string GMA()
        {
            var gid = SystemInfo.graphicsDeviceID;
            var gvid = SystemInfo.graphicsDeviceVendor;
            var vf = SystemInfo.graphicsDeviceVendorID;
            return string.Format("{0}_{1}_{2}", gid, gvid, vf).GetHashCode().ToString();
        }

        private bool bs(string pd)
        {
            string s = Location.r("_lvph") + GMA();
            string tmp = MDPWd(s);
            return pd == tmp;
        }

        private bool bsa(string pd)
        {
            string s = Location.r("_lvph") + GMAAP();
            string tmp = MDPWd(s);
            return pd == tmp;
        }

        private string GMAAP()
        {
            var pp = GMAA();
            var ppa = pp.Split('_');
            string s = ppa[0] + "_" + ppa[1];
            return s;
        }

        private string MDPWd(string rpwd)
        {
            string Ret;

            MD5 md5 = new MD5CryptoServiceProvider();
            byte[] palindata = Encoding.Default.GetBytes(rpwd);
            byte[] encryptdata = md5.ComputeHash(palindata);
            Ret = Convert.ToBase64String(encryptdata);
            return Ret;
        }

        private void OnDestroy()
        {
            if (hardResetBtn != null)
            {
                hardResetBtn.onClick.RemoveListener(UnloadBundle);
            }

            if (vbs.Count > 0)
            {
                for (int i = 0; i < vbs.Count; i++)
                {
                    if (vbs[i] != null)
                        Destroy(vbs[i]);
                }
            }

            if (lightSphereParents.Count > 0)
            {
                for (int i = 0; i < lightSphereParents.Count; i++)
                {
                    if (lightSphereParents[i] != null)
                    Destroy(lightSphereParents[i]);
                }
            }
            Utils.DestroyUI();
        }
    }
}
