﻿
using System.Collections;
using UnityEngine;

namespace PluginIdea.VolumeLight
{
    public class VolumetricLightMono : MonoBehaviour
    {
        public bool useLightColor = false;
        public Color mainColor = Color.white;

        public int numOfSides = 40;

        public float intensityFactor = 1;
        public float lightIntensity = 1;

        public float internalAlpha = 1;
        public float externalAlpha = 1;
        public float alphaScale = 1f;

        public bool deriveAngleFromLight = false;
        public float spotlightAngle = 35;
        public float startConeRadius = 0.01f;

        public int customGeometrySides = 30;
        public int customGeometrySegments = 10;

        public bool deriveAlphaFromIntensity = false;
        public bool deriveFadeEndFromLight = false;
        public float lightFadeStart = 0;
        public float lightFadeEnd = 3;

        public float noiseStrength = 0.5f;
        public float noiseScale = 0.5f;
        public Vector3 noiseSpeed = Vector3.one;

        public float endConeRadius => lightFadeEnd * Mathf.Tan(spotlightAngle * 0.0085f);

        public bool isUpdatePropertyFromLight;

        private VolumeLightGeometry Geometry = null;

        private Light _temporarySpotLight = null;

        private Camera MainCamera;

        private void Start()
        {
            MainCamera = Utils.CenterCamera;
            CreateGeometry();
        }

        public bool IsEnableLight
        {
            set { Geometry.Renderer.enabled = value; }
        }

        private void OnEnable()
        {
            if (Geometry != null) Geometry.Renderer.enabled = true;
        }

        private void OnDisable()
        {
            if (Geometry != null) Geometry.Renderer.enabled = false;
        }

        private void OnDestroy()
        {
            if (Geometry != null) DestroyImmediate(Geometry.gameObject);
            Geometry = null;
        }

        public void CreateGeometry()
        {
            UpdatePropertysFromLight();

            if (Geometry == null)
            {
                var lightGo = new GameObject("GeoMesh");
                Geometry = Utils.GetSafeComponent<VolumeLightGeometry>(lightGo);
                Geometry.AppendEssensialComponents(this);
            }

            Geometry.RegenerateMesh(numOfSides);
        }

        public void RefreshProperties()
        {
            UpdatePropertysFromLight();
            if (Geometry != null) Geometry.SetMaterialPropertys();
        }

        private void UpdatePropertysFromLight()
        {
            var spotLight = GetSpotLight();
            if (spotLight != null)
            {
                if (deriveFadeEndFromLight) lightFadeEnd = spotLight.range;
                if (deriveAlphaFromIntensity)
                {
                    internalAlpha = externalAlpha = spotLight.intensity * alphaScale;
                }
                if (deriveAngleFromLight) spotlightAngle = spotLight.spotAngle;
                if (useLightColor)
                {
                    mainColor = spotLight.color;
                }
            }
        }

        private Light GetSpotLight()
        {
            if (_temporarySpotLight == null) _temporarySpotLight = GetComponent<Light>();
            if (_temporarySpotLight != null && _temporarySpotLight.type == LightType.Spot) return _temporarySpotLight;
            return null;
        }

        private void Update()
        {
            if (isUpdatePropertyFromLight)
            {
                RefreshProperties();
            }
        }
    }
}